/*  igord.c --- broadcaster daemon

 Copyright (C) 2009 by Walter C. Pelissero

 Author: Walter C. Pelissero <walter@pelissero.de>
 Project: igorplug

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2, or (at
your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.*/

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "igorplug.h"
#include "util.h"

#define DEFAULT_HOST "localhost"
#define DEFAULT_PORT 5253


static int
open_socket ()
{
  int s;

  if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
      fprintf( stderr, "can't create socket (errno=%d)\n", errno);
      exit(2);
    }
  return s;
}

static struct in_addr
get_host_address (const char *host)
{
  struct hostent *he;
  struct in_addr ip;

  if (inet_aton(host, &ip))
    return ip;
  else if ((he = gethostbyname(host)) == NULL)
    {
      fprintf(stderr, "gethostbyname failed (errno=%d)\n", errno);
      exit(2);
    }
  return *(struct in_addr *)he->h_addr;
}

static int
get_service_port (const char *service)
{
  struct servent *se;

  if ((se = getservbyname(service, "udp")) == NULL)
    return atoi(service);
  return se->s_port;
}

static int
broadcast_packet (const unsigned char *buf, int socket, struct sockaddr_in destination)
{
  return sendto(socket, buf, IGORPLUG_DATA_LENGTH(buf) + IGORPLUG_DATA(0), 0,
		(struct sockaddr *)&destination, sizeof(destination));
}

static int
main_loop2 (int fd, unsigned tolerance, int socket, struct sockaddr_in destination)
{
  unsigned char *buf = igorplug_alloc_buffer();

  while (1)
    {
      if (igorplug_read_buffer2(fd, buf, tolerance) < 0)
	{
	  fprintf(stderr, "igorplug_read_buffer2 failed (errno=%d)\n", errno);
	  exit(2);
	}
      if (debug)
	print_packet(buf, stdout);
      broadcast_packet(buf, socket, destination);
    }
}

static void
usage (const char *my_name)
{
  printf("usage: %s [-d][-t tolerance] /dev/ugenX [host [port]]\n", my_name);
  printf("Broadcast IR packets via UDP\nhost defaults to \"%s\" and port defaults to %d\n",
	 DEFAULT_HOST, DEFAULT_PORT);
  exit(1);
}

int
main (int argc, char *argv[])
{
  int fd;
  unsigned tolerance = DEFAULT_TOLERANCE;
  struct sockaddr_in destination;
  int socket = -1;
  char *device = NULL;
  char c;

  destination.sin_family = AF_INET;
  destination.sin_port = htons(DEFAULT_PORT);
  destination.sin_addr = get_host_address(DEFAULT_HOST);
  memset(&(destination.sin_zero), '\0', 8);

  while ((c = getopt(argc, argv, "dt:")) != -1)
    {
      switch (c)
	{
	case 'd':
	  debug = true;
	  break;
	case 't':
	  tolerance = atoi(optarg);
	  break;
	case '?':
	default:
	  usage(argv[0]);
	}
    }
  if (argc - optind < 1)
    usage(argv[0]);
  device = argv[optind++];
  if (argc > optind)
   destination.sin_addr = get_host_address(argv[optind++]);
  if (argc > optind)
   destination.sin_port = get_service_port(argv[optind++]);
  if (argc > optind)
    usage(argv[0]);
  dprint("opening %s\n", device);
  fd = igorplug_open_device(device);
  socket = open_socket();
  main_loop2(fd, tolerance, socket, destination);
  /* should never get here */
  return 2;
}
