/*  xigor.c --- forward IR packets to X server

 Copyright (C) 2009 by Walter C. Pelissero

 Author: Walter C. Pelissero <walter@pelissero.de>
 Project: igorplug

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2, or (at
your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.*/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>

#include "igorplug.h"
#include "util.h"

#define DEFAULT_CONFIG_FILE "igorplug.x-events"

static void
setup_event (XKeyEvent *event, Display *display, Window window, Window root, KeySym key, int modifiers)
{
  event->display = display;
  event->window = window;
  event->root = root;
  event->serial = 0;
  event->send_event = 0;
  event->subwindow = None;
  event->time = CurrentTime;
  event->x = event->y = event->x_root = event->y_root = 0;
  event->same_screen = True;
  event->keycode = XKeysymToKeycode(display, key);
  event->state = modifiers;
  dprint("create_event: display=0x%X window=0x%X root=0x%X keysym=%d keycode=%d modifiers=%d -> event=0x%X\n",
	 display, window, root, key, event->keycode, modifiers, event);
}

static Window
current_window (Display *display)
{
  Window window;
  int revert;

  XGetInputFocus(display, &window, &revert);
  return window;
}

static void
send_event (XKeyEvent *event)
{
  if (!XSendEvent(event->display, event->window, True, KeyPressMask, (XEvent *)event))
    {
      fprintf(stderr, "XSendEvent() failed\n");
      exit(2);
    }
  XFlush(event->display);
}

static void
inject_event (Display *display, Window window, KeySym key)
{
  Window root = DefaultRootWindow(display);
  XKeyEvent event;

  if (window == None)
    window = current_window(display);
  setup_event(&event, display, window, root, key, 0);
  dprint("injecting %d\n", (unsigned int) key);
  /* key pressed */
  event.type = KeyPress;
  send_event(&event);
  /* key released */
  event.type = KeyRelease;
  send_event(&event);
}

static void *
parse_key (char *string)
{
  KeySym key;

  key = XStringToKeysym(string);
  if (key != NoSymbol)
    {
      KeySym *k = malloc(sizeof(KeySym));

      dprint("key: %s -> %ld\n", string, (unsigned long) key);
      *k = key;
      return k;
    }
  dprint("invalid X key \"%s\"\n", string);
  return 0;
}

static void
grab_pointer (Display *display)
{
  Cursor cursor = XCreateFontCursor(display, XC_crosshair);
  Window root = DefaultRootWindow(display);

  if (!GrabSuccess == XGrabPointer(display, root, False,
				   (ButtonPressMask | ButtonReleaseMask), GrabModeSync,
				   GrabModeAsync, root, cursor, CurrentTime))
    {
      fprintf(stderr, "can't grab mouse\n");
      exit(2);
    }
}

static void
ungrab_pointer (Display *display)
{
  XUngrabPointer(display, CurrentTime);
}

static Window
select_window (Display *display)
{
  XEvent event;
  Window window = None;
  Window root = DefaultRootWindow(display);

  printf("Click on a window to select it...\n");
  grab_pointer(display);
  while (window == None) 
    {
      XAllowEvents(display, SyncPointer, CurrentTime);
      XWindowEvent(display, root, ButtonPressMask | ButtonReleaseMask, &event);
      if (event.type == ButtonPress)
	{
	  Window child;
	  int x, y;

	  if (event.xbutton.subwindow)
	    {
	      XTranslateCoordinates(display, event.xbutton.window,
				    event.xbutton.subwindow,
				    event.xbutton.x_root,
				    event.xbutton.y_root,
				    &x, &y, &child);
	      if (child)
		window = child;
	      else
		window = event.xbutton.subwindow;
	    }
	  else
	    window = event.xbutton.window;
	}
    }
  ungrab_pointer(display);
  XFlush(display);
  printf("selected 0x%lX\n", (unsigned long) window);
  return window;
}

typedef struct 
{
  Display *display;
  Window window;
} destination;

static void
executor (KeySym *key, destination *dest)
{
  dprint("key=0x%X\n", *key);
  inject_event(dest->display, dest->window, *key);
}

static void
usage (const char *my_name)
{
  printf("usage: %s [-d][-f config_file][-w window_id][-t tolerance] /dev/ugenX\n\n", my_name);
  printf("if window_id = \"select\", choose the window interactively\n\
if window_id = \"current\", use the window that is in focus\n\
   at the start of the program\n\
otherwise window_id = should be a hexadecimal number\n\
if -w is not specified, use the window currently in focus at\n\
   the moment of each key press\n");
  exit(1);
}

int
main (int argc, char *argv[])
{
  int fd;
  destination dest;
  unsigned tolerance = DEFAULT_TOLERANCE;
  char c;
  char *config_file = DEFAULT_CONFIG_FILE;
  action_entry *actions;

  dest.display = XOpenDisplay(0);
  dest.window = None;

  if (!dest.display)
    {
      fprintf(stderr, "Can't open display.");
      return 1;
    }
  while ((c = getopt(argc, argv, "dw:t:f:")) != -1)
    {
      switch (c)
	{
	case 'd':
	  debug = true;
	  break;
	case 'w':
	  if (strcmp(optarg, "select") == 0)
	    dest.window = select_window(dest.display);
	  else if (strcmp(optarg, "current") == 0)
	    dest.window = current_window(dest.display);
	  else
	    dest.window = strtol(optarg, 0, 16);
	  break;
	case 'f':
	  config_file = optarg;
	  break;
	case 't':
	  tolerance = atoi(optarg);
	  break;
	case '?':
	default:
	  usage(argv[0]);
	}
    }
  if (argc == optind)
    usage(argv[0]);
  actions = read_config_file(config_file, parse_key);
  if (!actions)
    {
      fprintf(stderr, "No actions in config file.\n");
      exit(2);
    }
  dprint("opening %s\n", argv[optind]);
  fd = igorplug_open_device(argv[optind]);
  dprint("window=0x%X\n", (unsigned) dest.window);
  main_loop(fd, (action_executor) executor, actions, tolerance, &dest);
  /* shouldn't reach this */
  XCloseDisplay(dest.display);
  return 2;
}
