/*  igor_pty.c --- inject igorplug events to a program through a pty

 Copyright (C) 2009 by Walter C. Pelissero

 Author: Walter C. Pelissero <walter@pelissero.de>
 Project: igorplug

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2, or (at
your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.*/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <libutil.h>
#include <signal.h>

#include "igorplug.h"
#include "util.h"

#define DEFAULT_CONFIG_FILE "igorplug.keystrokes"

static int
isoctal (char c)
{
  return c >= '0' && c <= '7';
}

static unsigned char
parse_octal (char **string)
{
  char *p = *string;
  unsigned char x = 0;
  int i;

  for (i = 0; i < 3 && *p; ++i, ++p)
    {
      if (!isoctal(*p))
	break;
      x = x * 8 + *p - '0';
    }
  *string = p;
  return x;
}

static void
unescape_string (char *string)
{
  char *p = string;

  while (*string)
    {
      if (*string == '\\')
	{
	  ++string;
	  switch (*string)
	    {
	    case 'n':
	      *p++ = '\n';
	      ++string;
	      break;
	    case 't':
	      *p++ = '\t';
	      ++string;
	      break;
	    case 'r':
	      *p++ = '\r';
	      ++string;
	      break;
	    case '0':
	      ++string;
	      *p++ = parse_octal(&string);
	      break;
	    default:
	      *p++ = *string++;
	      break;
	    }
	}
      else
	*p++ = *string++;
    }
  *p = '\0';
}

static void *
parse_string (char *string)
{
  int len = strlen(string);

  if (len > 1 && string[0] == '"' && string[len - 1] == '"')
    {
      char *stripped;

      stripped = malloc(len - 1);
      memcpy(stripped, string + 1, len - 2);
      stripped[len - 1] = '\0';
      unescape_string(stripped);
      return stripped;
    }
  unescape_string(string);
  return strdup(string);
}

static void
executor (char *string, int *pty_fd)
{
  int n = strlen(string);

  dprint("string=%s\n", string);
  if (write(*pty_fd, string, n) < n)
    exit(1);
}

static void
usage (const char *my_name)
{
  printf("usage: %s [-d][-f config_file][-t tolerance][-e][-i|I] /dev/ugenX command [arg ...]\n", my_name);
  exit(1);
}

static void
copy_loop (int in, int out)
{
  char buf[1024];
  int n;

  while ((n = read(in, buf, sizeof(buf))) > 0)
    {
      char *p = buf;
      int o;

      while (n)
	{
	  o = write(out, p, n);
	  if (o < 0)
	    return;
	  n -= o;
	  p += o;
	}
    }
}

static struct termios orig_termios;
int orig_fd;

static void
tty_restore (int fd)
{
  if (tcsetattr(fd, TCSAFLUSH, &orig_termios) < 0)
    exit(2);
}

static void
sigterm_handler (int sig)
{
  tty_restore(orig_fd);
}

#if 0
static void
propagate_signal (int sig)
{
  kill(0, sig);
}
#endif

static pid_t
pty_fork (int *fd, int interactive)
{
  struct winsize size;

  orig_fd = STDIN_FILENO;
  if (tcgetattr(STDIN_FILENO, &orig_termios) < 0)
    {
      fprintf(stderr, "can't tcgetattr\n");
      exit(2);
    }
  if (interactive)
    {
      signal(SIGTERM, sigterm_handler);
      if (ioctl(STDIN_FILENO, TIOCGWINSZ, (char *) &size) < 0)
	{
	  fprintf(stderr, "can't TIOCGWINSZ\n");
	  exit(2);
	}
      return forkpty(fd, 0, &orig_termios, &size);
    }
  else
    return forkpty(fd, 0, 0, 0);
}

static void
tty_noecho (int fd)
{
  struct termios tp;

  orig_fd = fd;
  if (tcgetattr(STDIN_FILENO, &tp) < 0)
    {
      fprintf(stderr, "can't tcgetattr\n");
      exit(2);
    }
  tp.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
  tp.c_oflag &= ~ONLCR;
  if (tcsetattr(fd, TCSANOW, &tp) < 0)
    exit(1);
}

static void
tty_raw (int fd)
{
  struct termios tp;

  orig_fd = fd;
  if (tcgetattr(fd, &tp) < 0)
    exit(2);
  tp.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
  tp.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
  tp.c_cflag &= ~(CSIZE | PARENB);
  tp.c_cflag |= CS8;
  tp.c_oflag &= ~OPOST;
  tp.c_cc[VMIN] = 1;
  tp.c_cc[VTIME] = 0;
  if (tcsetattr(fd, TCSAFLUSH, &tp) < 0)
    exit(2);
}

static void
restore_tty_on_exit ()
{
  tty_restore(orig_fd);
}
static int pty_fd;

static void
sigwinch_handler (int sig)
{
  struct winsize size;

  ioctl(STDIN_FILENO, TIOCGWINSZ, (char *) &size);
  ioctl(pty_fd, TIOCSWINSZ, (char *) &size);
}

int
main (int argc, char *argv[])
{
  pid_t pid;
  unsigned tolerance = DEFAULT_TOLERANCE;
  char c;
  char *config_file = DEFAULT_CONFIG_FILE;
  action_entry *actions;
  int noecho = 0;
  int interactive = isatty(STDIN_FILENO);

  while ((c = getopt(argc, argv, "dt:f:eiI")) != -1)
    {
      switch (c)
	{
	case 'd':
	  debug = true;
	  break;
	case 'f':
	  config_file = optarg;
	  break;
	case 't':
	  tolerance = atoi(optarg);
	  break;
	case 'e':
	  noecho = 1;
	  break;
	case 'i':
	  interactive = 1;
	  break;
	case 'I':
	  interactive = 0;
	  break;
	case '?':
	default:
	  usage(argv[0]);
	}
    }
  if ((argc - optind) < 2)
    usage(argv[0]);
  actions = read_config_file(config_file, parse_string);
  if (!actions)
    {
      fprintf(stderr, "No actions in config file.\n");
      exit(2);
    }
  pid = pty_fork(&pty_fd, interactive);
  if (pid < 0)
    fprintf(stderr, "can't fork\n");
  else if (pid == 0)		/* child (the controlled program) */
    {
      if (noecho)
	tty_noecho(STDIN_FILENO);
      ++optind;
      execvp(argv[optind], argv + optind);
      fprintf(stderr, "can't exec %s\n", argv[optind]);
      kill(getppid(), SIGTERM);
    }
  else				/* parent */
    {
      pid_t ir_monitor_pid = fork();

      if (ir_monitor_pid < 0)
	fprintf(stderr, "can't fork\n");
      else if (ir_monitor_pid == 0) /* child (the IR event loop) */
	{
	  int fd;

	  dprint("opening %s\n", argv[optind]);
	  fd = igorplug_open_device(argv[optind]);
	  if (fd >= 0)
	    main_loop(fd, (action_executor) executor, actions, tolerance, &pty_fd);
	  else
	    fprintf(stderr, "Can't open IR receiver \"%s\"\r\n", argv[optind]);
	  kill(getppid(), SIGTERM);
	}
      else			/* parent */
	{
	  if (interactive)
	    {
	      tty_raw(STDIN_FILENO);
	      atexit(restore_tty_on_exit);
	    }
	  signal(SIGTERM, sigterm_handler);
	  {
	    
	    pid_t pid = fork();

	    if (pid < 0)
	      fprintf(stderr, "can't fork\n");
	    else if (pid == 0)	/* child (the output echo) */
	      {
		copy_loop(pty_fd, STDOUT_FILENO);
		kill(getppid(), SIGTERM);
	      }
	    else		/* parent (the input echo) */
	      {
		if (interactive)
		  signal(SIGWINCH, sigwinch_handler);
		copy_loop(STDIN_FILENO, pty_fd);
		kill(pid, SIGTERM);
	      }
	    kill(ir_monitor_pid, SIGTERM);
	  }
	}
      kill(pid, SIGTERM);
    }
  return 2;
}
