 ;;; ffi.lisp --- igorplug UFFI interface

 ;;; Copyright (C) 2009 by Walter C. Pelissero

 ;;; Author: Walter C. Pelissero <walter@pelissero.de>
 ;;; Project: igorplug

#+cmu (ext:file-comment "$Module: ffi.lisp, Time-stamp: <2009-03-26 13:29:33 wcp> $")

;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1
;;; of the License, or (at your option) any later version.
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with this library; if not, write to the Free
;;; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
;;; 02111-1307 USA

(in-package :igor-plug)

(defconstant +max-packet-length+ 36)
(defconstant +header-length+ 3)
(defconstant +buffer-length+ 255
  "Length of the i/o buffer.  In reality this could be much smaller because the buffer on the device is 36 bytes.")

(defconstant +packet-length-position+ 0
  "Position of the length byte in the packet.")

(defconstant +packet-serial-position+ 1
  "Position of the serial number byte in the packet.")

(defconstant +packet-offset-position+ 2
  "Position of the offset byte in the packet.")

(defconstant +header-length+ 3)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(eval-when (:compile-toplevel :load-toplevel)
  (uffi:load-foreign-library (make-pathname :defaults #.(or *compile-file-pathname*
							    *load-pathname*)
					    :name "libigor"
					    :type "so")
			     :module "igorplug"))

(uffi:def-array-pointer igorplug-buffer :unsigned-byte)

(uffi:def-function ("igorplug_alloc_buffer" %igorplug-alloc-buffer) 
  ()
  :returning igorplug-buffer
  :module "igorplug")

;;
;; int igorplug_open_device(char *device_pathname)
;;
(uffi:def-function ("igorplug_open_device" %igorplug-open-device) 
  ((device_pathname :cstring))
  :returning :int
  :module "igorplug")

;;
;; int igorplug_read_buffer(int fd, unsigned char *buf)
;;
(uffi:def-function ("igorplug_read_buffer" %igorplug-read-buffer) 
  ((fd :int) (buf igorplug-buffer))
  :returning :int
  :module "igorplug")

;;
;; int igorplug_receive_code(int fd, unsigned char *buf, unsigned tolerance)
;;
(uffi:def-function ("igorplug_receive_code" %igorplug-receive-code) 
  ((fd :int) (buf igorplug-buffer) (tolerance :unsigned-int))
  :returning :int
  :module "igorplug")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-condition igorplug-error (error)
  ((errno :initarg :errno :reader igorplug-error-no))
  (:report (lambda (condition stream)
	     (format stream "errno=~D" (igorplug-error-no condition)))))


(defun alloc-buffer ()
  (let ((buffer (%igorplug-alloc-buffer)))
    (when (uffi:null-pointer-p buffer)
      (error 'igorplug-error :errno (sb-alien:get-errno)))
    buffer))

(deftype ubyte ()
  `(unsigned-byte 8))

(deftype igorplug-packet ()
  `(vector ubyte))

(defstruct (igorplug-device
	     (:conc-name device-))
  pathname
  fd
  (buffer (alloc-buffer)))

(defun close-device (device)
  (declare (type igorplug-device device))
  (sb-posix:close (device-fd device))
  (uffi:free-foreign-object (device-buffer device))
  (setf (device-fd device) nil
	(device-buffer device) nil))

(defun open-device (pathname)
  (setf pathname (merge-pathnames pathname #P"/dev/"))
  (let ((fd (%igorplug-open-device (namestring pathname))))
    (when (< fd 0)
      (error 'igorplug-error :errno (sb-alien:get-errno)))
    (make-igorplug-device :pathname pathname
			  :fd fd)))

(defmacro with-device ((device pathname) &body forms)
  `(let ((,device (open-device ,pathname)))
     (unwind-protect (progn ,@forms)
       (close-device ,device))))

(declaim (inline buffer-ref packet-length packet-serial packet-offset))

(defun buffer-ref (buffer i)
  (uffi:deref-array buffer igorplug-buffer i))

(defun packet-length (buffer)
  (buffer-ref buffer +packet-length-position+))

(defun packet-serial (buffer)
  (buffer-ref buffer +packet-serial-position+))

(defun packet-offset (buffer)
  (buffer-ref buffer +packet-offset-position+))

(defun read-packet (device &optional buffer)
  (unless buffer
    (setf buffer (device-buffer device)))
  (loop
     with len = 0 and seq = -2
     for ret = (%igorplug-receive-code (device-fd device) buffer 1)
     when (< ret 0)
     do (error 'igorplug-error :errno (sb-alien:get-errno))
     ;; Loop until we have two consecutive packets of the same length.
     ;; This should giv use enough confidence we are not receiving
     ;; random garbage.
     until (and (= len (packet-length buffer))
		(= (1+ seq) (packet-serial buffer)))
     do (setf len (packet-length buffer)
	      seq (packet-serial buffer)))
  (values))

(defun print-buffer (device &optional (stream *standard-output*))
  (let ((buffer (device-buffer device)))
    (format stream "#~3D ~3D:"
	    (packet-serial buffer)
	    (packet-offset buffer)
	    (packet-length buffer))
    (loop
       for i from 0 below (packet-length buffer)
       do (format stream " ~2,'0X"
		  (uffi:deref-array buffer igorplug-buffer
				    (+ i +header-length+))))
    (terpri stream)))

(defun buffer-to-vector (buffer)
  (loop
     with len = (packet-length buffer)
     with v = (make-sequence 'igorplug-packet len)
     for i from 0 below len
     do (setf (aref v i) (buffer-ref buffer (+ i +header-length+)))
     finally (return v)))

(defun echo-packets (device-pathname &optional (stream *standard-output*))
  (with-device (device device-pathname)
    (loop
       (read-packet device)
       (print-buffer device stream))))
